
/* Erstellt: 2020-05-22
 * Autor:  Roland Kaiser, OE2ROL
 * Vielen Dank auch an Balazs Kelemen der die Library LcdBarGraph.h erstellt hat.
 * Version V1.9
 * geschrieben für Arduino Nano ATmega 328P(Old Bootloader)
 * Der Autor übernimmt keine Haftung etwaiger Folgeschäden an Transceivern oder dgl.
 * 
 * Diese Software ist für den eigenen privaten Gebrauch frei verfügbar, jedoch nicht für kommerzielle Zwecke ohne Genehmigung des Autors.
 * Veränderungen und Verbesserungen der Software bitte wieder an den Autor Roland Kaiser OE2ROL (roland.kaiser@sbg.at) senden.
 *
 * Funktionsweise:
 * 
 * Die Software QO-100 Transvertersteuerung wurde am ICOM IC820H getestet sollte auch für IC821H und auch für IC-9700 passen. 
 * Die 2m Frequenz zum Upconverter wird an der CAT-Schnittstelle des Transceivers gelesen
 * Frequenzausgabe der umgerechneten Frequenzen (TX -> Eingangsfrequenz + Oszillatorfrequenz des Upconverters) ( RX -> TX Frequenz + Offsetfrequenz des Satelliten) und Ausgabe der RX un TX Freq. auf ein LCD Display 
 * Ausgabe der vor- und rücklaufenden Leistung mit Bargraphanzeige auf dem LCD Display 
 * Temperatursteuerung des Lüfters zur Kühlung der PA 
 * Ausgabe der Frequenz und Betriebsart die an die CAT-Schnittstelle übertragen wurde, wird über eine RS232 Schnittstelle ausgegeben. Für Logprogramme die über OmniRig arbeiten z.B. Log4OM
 */

#include <LiquidCrystal.h>
#include <LcdBarGraph.h>
#include <SoftwareSerial.h>
#include <AltSoftSerial.h>

byte lcdNumCols = 20; // LCD Anzahl der Stellen
byte FWD = 0;         // Analogeingang 0 FWD
byte REF = 1;         // Analogeingang 1 REF
int Temp = 2;         // Analogeingang 2 Temp
int pwmPinBUZ71 = 3;  // PWM Ausgang Lüftersteuerung
int temp;             // Variable für Temperatur
int tempPin = 2;      // Temperatursensoreingang

LiquidCrystal lcd(7, 10, 5, 4, 6, 2); // LCD Anschlussbelegung
LcdBarGraph fwd(&lcd, 16, 22, 2);    // erstellt 16 char bargraph 4.Stelle, 3.Zeile (Stelle 2)
LcdBarGraph ref(&lcd, 16, 22, 3);    // erstellt 16 char bargraph 4.Stelle, 4.Zeile (Stelle 3)
SoftwareSerial CAT(11, 12);          // RX, TX Pin CAT Schnittstelle
AltSoftSerial COM1(8, 9);            // RX, TX RS232 (COM1) Schnittstelle

#define BROADCAST_ADDRESS    0x00 //Broadcast address
#define CONTROLLER_ADDRESS   0xE0 //Controller address

#define START_BYTE           0xFE //Start byte
#define STOP_BYTE            0xFD //Stop byte

#define CMD_TRANS_FREQ       0x00 //Transfers operating frequency data
#define CMD_TRANS_MODE       0x01 //Transfers operating mode data

#define CMD_READ_FREQ        0x03 //Read operating frequency data
#define CMD_READ_MODE        0x04 //Read operating mode data

#define CMD_WRITE_FREQ       0x05 //Write operating frequency data
#define CMD_WRITE_MODE       0x06 //Write operating mode data

const uint32_t decMulti[]    = {1000000000, 100000000, 10000000, 1000000, 100000, 10000, 1000, 100, 10, 1};
const uint16_t baudRates[]   = {19200, 9600, 4800, 1200};

uint8_t  radio_address;         //Tranceiver address
uint16_t  baud_rate;            //Current baud speed
uint32_t  readtimeout;          //Serial port read timeout
uint8_t  read_buffer[12];       //Read buffer CAT
uint32_t  frequency;            //Current frequency in Hz
const char* mode[] = {"LSB","USB","AM","CW","FSK","FM","WFM"};
uint32_t TXfrequency;           //umgerechnete Sendefrequenz
uint32_t TXosz;                 //TX Oszillatorfrequenz
uint32_t RXfrequency;           //umgerechnete Empfangsfrequenz
uint32_t RXosz;                 //RX Oszillatorfrequenz
uint8_t read_buffer_freq[12];   //CAT Frequenz Read buffer
uint8_t read_buffer_mode[12];   //CAT Mode Read buffer
uint8_t COM1_in[7];             // COM1 in Array
int  serInIndx  = 0;            // index von COM1_in[] in welches das nächste eingehende byte geschrieben wird

void setup()
{  
  lcd.begin(2, lcdNumCols); // LCD init
  lcd.clear();              // LCD löschen
  delay(100);               // Verzögerung
  lcd.setCursor(0, 0);      // 1. Zeichen, 1. Zeile
  lcd.print("RX");          // schreiben
  lcd.setCursor(17, 0);     // 1. Zeichen, 1. Zeile
  lcd.print("MHz");         // schreiben
  lcd.setCursor(0, 1);      // 1. Zeichen, 1. Zeile
  lcd.print("TX");          // schreiben
  lcd.setCursor(17, 1);     // 1. Zeichen, 1. Zeile
  lcd.print("MHz");         // schreiben
  lcd.setCursor(18, 2);     // 1. Zeichen, 1. Zeile
  lcd.print("FWD");         // schreiben
  lcd.setCursor(18, 3);     // 1. Zeichen, 1. Zeile
  lcd.print("REF");         // schreiben
  
  pinMode(pwmPinBUZ71, OUTPUT);
  
  Serial.begin(9600); // Datenrate für HardwareSerial setzen
  COM1.begin(9600);   // Datenrate für AltSoftSerial (COM1) an Pin 8/9 setzen
  
    
  configRadioBaud(9600);
  radio_address = 0x00;

  while(radio_address == 0x00) {
    if (!searchRadio()) {
  
    } else {

      sendCatRequest(CMD_READ_FREQ);
      delay(100);
      sendCatRequest(CMD_READ_MODE);
      break;
    }
  }
}

void fan() {
  temp = analogRead(tempPin);
  int mv = ( temp/1024.0)*5000;
  int cel = mv/10;
  TCCR2B = TCCR2B & B11111000 | 0x01; // Timer 2 Teiler auf 1 setzen für PWM Frequenz 31300 Hz
  analogWrite(pwmPinBUZ71, mv/35);    // hier kann die Ansprechschwelle der Lüftersteuerung verändert werden (mv/xx)


  Serial.begin(9600);
  Serial.print("TEMP. = ");
  Serial.print(cel);
  Serial.print("°C");
  Serial.print(" ");
  Serial.print(mv/25);
  Serial.println();
 
  delay(100);
}

void Bargraph(void)
{
  fwd.drawValue( analogRead(FWD), 1024); // schreibe Bargraph aus dem gelesenen Analogwert 
  ref.drawValue( analogRead(REF), 1024); // schreibe Bargraph aus dem gelesenen Analogwert
  delay(100); // Verzögerung
}

void configRadioBaud(uint16_t  baudrate)
{
  if(baud_rate != baudrate){
    CAT.end();
    CAT.begin(baudrate);      //Datenrate für SoftwareSerial (CAT) an Pin 11/12 setzen
   
  }
  
  baud_rate = baudrate;

  switch (baud_rate) {
    case 4800:
       readtimeout = 50000;
       break;
    case 9600:
       readtimeout = 40000;
       break;
    case 19200:
       readtimeout = 30000;
       break;
    default:
       readtimeout = 100000;
       break;
  }
}

uint8_t readLine(void)
{
  uint8_t byte;
  uint8_t counter = 0;
  uint32_t ed = readtimeout;

  while (true)
  {
    while (!CAT.available()) {
      if (--ed == 0)return 0;
    }
    ed = readtimeout;
    byte = CAT.read();
    if (byte == 0xFF)continue; // TODO skip to start byte instead
    //Serial.begin(9600);      // opens serial port, sets data rate to 9600 bps
    //Serial.print(byte,HEX);

    read_buffer[counter++] = byte;
    if (STOP_BYTE == byte) break;

    if (counter >= sizeof(read_buffer))return 0;
  }
  return counter;
}

void sendCatRequest(uint8_t requestCode)
{
  uint8_t req[] = {START_BYTE, START_BYTE, radio_address, CONTROLLER_ADDRESS, requestCode, STOP_BYTE};

  for (uint8_t i = 0; i < sizeof(req); i++) {
    CAT.write(req[i]);
    }
}

bool searchRadio()
{
  for(uint8_t baud=0; baud<sizeof(baudRates); baud++){
  
    configRadioBaud(baudRates[baud]);
    sendCatRequest(CMD_READ_FREQ);
    
    if (readLine() > 0)
    {
      if(read_buffer[0]==START_BYTE && read_buffer[1]==START_BYTE){
        radio_address = read_buffer[3];
      }
      return true;
    }
  }
  radio_address = 0xFF;
  return false;
}

void printFrequency(void)
{
  frequency = 0;
  //FE FE E0 42 03 <00 90 42 44 01> FD = 0144429000 Hz
  for (uint8_t i = 0; i < 5; i++) {
    if(read_buffer[9 - i] == 0xFD)continue;
    
    frequency += (read_buffer[9 - i] >> 4) * decMulti[i * 2];         // Frequenz aus dem Buffer lesen und in Dezimal konvertieren
    frequency += (read_buffer[9 - i] & 0x0F) * decMulti[i * 2 + 1];   // Frequenz aus dem Buffer lesen und in Dezimal konvertieren
  }
    for (uint8_t i = 0; i < 11; i++){
    read_buffer_freq[i]=(read_buffer[i]);                             // Zeichen vom Buffer lesen und an Frequenzbuffer übergeben
    }
    read_buffer_freq[2]= CONTROLLER_ADDRESS;                          // setzen der Controlleradresse EO
    read_buffer_freq[4]= CMD_READ_FREQ;                               // setzen der Frequenz-Leseadresse 03

TXosz=2256000000;                                                     //TX Oszillatorfrequenz [Hz]
TXfrequency=(frequency)+(TXosz);                                      //Berechnung der Ausgangsfrequenz (Eingangsfreq. + Oszillatorfreq.)
lcd.setCursor(4, 1);
lcd.print(TXfrequency/1000000000%10); //Tx 1G Stelle ausgeben
lcd.setCursor(5, 1);
lcd.print(TXfrequency/100000000%10);  //Tx 100M  Stelle ausgeben
lcd.setCursor(6, 1);
lcd.print(TXfrequency/10000000%10);   //Tx 10M Stelle ausgeben
lcd.setCursor(7, 1);
lcd.print(TXfrequency/1000000%10);    //Tx 1M Stelle ausgeben
lcd.setCursor(8, 1);
lcd.print(",");                       //, ausgeben
lcd.setCursor(9, 1);
lcd.print(TXfrequency/100000%10);     //Tx 100k Stelle ausgeben
lcd.setCursor(10, 1);
lcd.print(TXfrequency/10000%10);      //Tx 10k Stelle ausgeben
lcd.setCursor(11, 1);
lcd.print(TXfrequency/1000%10);       //Tx 1k Stelle ausgeben
lcd.setCursor(12, 1);
lcd.print(TXfrequency/100%10);        //Tx 100Hz Stelle ausgeben
lcd.setCursor(13, 1);
lcd.print(TXfrequency/10%10);         //Tx 10Hz Stelle ausgeben
lcd.setCursor(14, 1);
lcd.print(TXfrequency%10);            //Tx 1Hz Stelle ausgeben

RXfrequency=((TXfrequency/1000)+8089500);                             // setzen RX-TX shift 8,089.500GHz = Diff Uplink und Downlink QO100
lcd.setCursor(1, 2);
lcd.print(RXfrequency/10000000%10);   //Rx 10G Stelle ausgeben
lcd.setCursor(2, 2);
lcd.print(RXfrequency/1000000%10);    //Rx 1G Stelle ausgeben
lcd.setCursor(3, 2);
lcd.print(RXfrequency/100000%10);     //Rx 100M Stelle ausgeben
lcd.setCursor(4, 2);
lcd.print(RXfrequency/10000%10);      //Rx 10M Stelle ausgeben
lcd.setCursor(5, 2);
lcd.print(RXfrequency/1000%10);       //Rx 1M Stelle ausgeben
lcd.setCursor(6, 2);
lcd.print(",");                       //, ausgeben
lcd.setCursor(7, 2);
lcd.print(RXfrequency/100%10);        //Rx 100k Stelle ausgeben
lcd.setCursor(8, 2);
lcd.print(RXfrequency/10%10);         //Rx 10k Stelle ausgeben
lcd.setCursor(9, 2);
lcd.print(RXfrequency%10);            //Rx 1k Stelle ausgeben
lcd.setCursor(10, 2);
lcd.print(TXfrequency/100%10);        //Tx 100Hz Stelle ausgeben
lcd.setCursor(11, 2);
lcd.print(TXfrequency/10%10);         //Tx 10Hz Stelle ausgeben
lcd.setCursor(12, 2);
lcd.print(TXfrequency%10);            //Tx 1Hz Stelle ausgeben

/*Serial.begin(9600);
  Serial.print("FREQ = ");
  Serial.print(RXfrequency);
  Serial.print("Hz");
  //Serial.print(" ");
  Serial.println();
  */
}

void rs_232read()
{
uint8_t sb;   
    if(COM1.available()>0){                     //Abfrage ob Daten an der RS232 Schnittstelle ankommen
       while (COM1.available()){                //wenn Daten an der RS232 Schnittstelle angekommen sind, dann
          sb = COM1.read();                     //Daten in Variable einlesen  
          COM1_in[serInIndx] = sb;              //Daten der Variablen in ein Array übertragen
          serInIndx++;                          //Arrayindex hochzählen
       }
    serInIndx  = 0;                             //Arrayindex zurücksetzen
    } 
}
 
void rs_232send_mode()
{             
     if (COM1_in[4] == CMD_READ_MODE){          //wenn an der 5. Stelle das HEX Kommando "CMD_READ_MODE" enthalten ist, dann
         for (uint8_t i = 0; i < 6; i++){       //Zählschleife für den Arrayindex ausführen
         COM1.write(COM1_in[i]);                //Empfangenen Code für READ MODE an RS232 senden
         }
           for (uint8_t j = 0; j < 8; j++){     //Zählschleife für den Arrayindex ausführen
           COM1.write(read_buffer_mode[j]);     //Code vom Mode Buffer an rs232 senden
           memset(COM1_in,0,sizeof(COM1_in));   //Array des RS232 Lesespeichers löschen
           }
      }
}

void rs_232send_freq()
{  
     if (COM1_in[4] == CMD_READ_FREQ){          //wenn an der 5. Stelle das HEX Kommando "CMD_READ_FREQ" enthalten ist, dann
         for (uint8_t k = 0; k < 6; k++){       //Zählschleife für den Arrayindex ausführen
         COM1.write(COM1_in[k]);                //Empfangenen Code für READ MODE an RS232 senden
           }
           for (uint8_t l = 0; l < 11; l++){    //Zählschleife für den Arrayindex ausführen
           COM1.write(read_buffer_freq[l]);     //Code vom Frequenz Buffer an rs232 senden
           memset(COM1_in,0,sizeof(COM1_in));   //Array des RS232 Lesespeichers löschen
           }
     } 
}  

void printMode(void)
{
  //FE FE E0 00 04 <00 01> FD  = USB
  //read_buffer[6] -> 01 - Wide, 02 - Medium, 03 - Narrow
  for (uint8_t i = 0; i < 8; i++) 
  {  
  read_buffer_mode[i]=(read_buffer[i]);     // Zeichen vom Buffer lesen und an Modebuffer übergeben
  }
  read_buffer_mode[2]= CONTROLLER_ADDRESS;  // setzen der Controlleradresse EO
  read_buffer_mode[4]= CMD_READ_MODE;       // setzen der Mode-Leseadresse 04
}

void processCatMessages()
{
  /*
    <FE FE E0 42 04 <00 01> FD  - LSB
    <FE FE E0 42 03 <00 00 58 45 01> FD  -145.580.000

    FE FE - start bytes
    00/E0 - target address (broadcast/controller)
    42 - source address
    00/03 - data type
    <data>
    FD - stop byte
  */
 
  while (CAT.available()) {
    uint8_t knowncommand = 1;
    uint8_t r;
    if (readLine() > 0) {
      if (read_buffer[0] == START_BYTE && read_buffer[1] == START_BYTE) {
        if (read_buffer[3] == radio_address) {
          if (read_buffer[2] == BROADCAST_ADDRESS) {
            switch (read_buffer[4]) {
              case CMD_TRANS_FREQ:
                printFrequency();
                break;
              case CMD_TRANS_MODE:
                printMode();
                break;
              default:
                knowncommand = false;
            }
          } else if (read_buffer[2] == CONTROLLER_ADDRESS) {
            switch (read_buffer[4]) {
              case CMD_READ_FREQ:
                printFrequency();
                break;
              case CMD_READ_MODE:
                printMode();
                break;
              default:
                knowncommand = false;
            }
          }
        } else {
       
        }
      }
    }
  }
  for (uint8_t i = 0; i < sizeof(read_buffer); i++) 
    {
     if (read_buffer[i] == STOP_BYTE)break;
    }
}

void loop()
{
  fan();                  //Lüfter Programmschritt ausführen
  Bargraph();             //Bargraph Programmschritt ausführen
  processCatMessages();   //CAT Schnitttstelle Programmschritt ausführen
  rs_232read();           //RS232 lesen Programmschritt ausführen
  rs_232send_mode();      //senden der "Mode" Daten für RS232 Programmschritt ausführen
  rs_232send_freq();      //senden der "Frequenz" Daten für RS232 Programmschritt ausführen
  
}
